/**
 * k-errors queries
 * 
 * Aligment using suffix filters, see README for references.
 */

#ifndef _IndelQuery_H_
#define _IndelQuery_H_

#include "Pattern.h"
#include "InputReader.h"
#include "OutputWriter.h"
#include "EditDistance.h"
#include "ResultSet.h"
#include "Query.h"

#include <string>
#include <cstdlib> // exit()
#include <cmath> // pow()

class IndelQuery : public Query
{
public:
    IndelQuery(TextCollection *tc, OutputWriter &ow, bool vrb, 
          bool incr, bool rvrs, unsigned rprt)
        : Query(tc, ow, vrb, incr, rvrs, rprt), results(0), psize(0), splitPos(0), ed(0)
    { 
        results = new ResultSet(tc->getLength());
    }

    virtual ~IndelQuery() 
    {
        delete results; results = 0;
	delete [] splitPos; splitPos = 0;
        delete ed;
        ed = 0;
    }

protected:
    void firstStep(void) 
    {
        results->clear();
        
        psize = patlen / (klimit + 1);
        calculateOptimalPartition();
        counterstart = callcounter;
        for (unsigned i = 1; i <= klimit+1; i++) {
            if (debug) std::cout << "new try: " << splitPos[i] << "\n";
            uchar *reversepat = new uchar[splitPos[i]+1];
            for (unsigned j = 0; j < splitPos[i]; j++)
                reversepat[j] = pat[splitPos[i]-j-1];
            reversepat[splitPos[i]] = 0;
            if (debug)
                std::cerr << "revpat: " << reversepat << std::endl;
//            ed = new EditDistance(reversepat,splitPos[i],splitPos[i]+i+1);
            ed = new MyersEditDistanceIncremental(reversepat,splitPos[i],splitPos[i]+i+1);
            nextStep(splitPos[i],i, -1, 1);
            delete [] reversepat;
            delete ed;
            ed = 0;
        }
    }

private:
    void nextStep(int pos, unsigned split, int maxk, unsigned count) 
    {
        callcounter++;
        if (callcounter - counterstart > rlimit) return;
        if (sum >= report) return;
        if ((int)ed->dist() <= maxk) { checkMatch(); }
        if (split && pos == (int)splitPos[split]) {
            maxk++;
            --split;
        }
        if (count > (int)ed->length + klimit) return;

        // ALPHABET has been defined in Query
        for (const char *cp = ALPHABET; cp < ALPHABET + ALPHABET_SIZE; ++cp) {
            char c = *cp;
            if (!pushChar(c)) continue;
            int colmin = ed->pushChar(c);
            if (debug)
                std::cerr << "c = " << c << ", pos = " << pos << ", split = " << split << ", maxk = " << maxk << ", colmin = " << colmin 
                          << ", dist = " << ed->dist() << std::endl;
            if (colmin <= maxk) nextStep(pos - 1, split, maxk, count + 1);
            if (debug) std::cerr << "pop c = " << c << std::endl;
            ed->popChar();
            popChar();
        }
    }

    inline void checkMatch(void) 
    {
        if (debug) std::cerr << "min,max: " << smin.top() << ", " << smax.top() << std::endl;
        ulong min = smin.top();
        ulong max = smax.top();
        MyersEditDistance med(patlen,patlen+klimit);
        for (ulong i = min; i <= max; i++) {
            if (sum >= report) break;
            if (!results->get(i-1))
            {
                results->set(i-1);
                med.init((uchar const *)pat);
                validateCandidate(med, i);
            }
        }
    }

    void validateCandidate(MyersEditDistance &med, ulong sapos) 
    {
/*        TextCollection::position_result pos;
        tc.getPosition(pos, sapos);
        uchar const * text = tc.getText(pos);
        uchar *candidate = new uchar[patlen+klimit+1];
        unsigned l = 0;
        while (l < patlen+klimit && text[l] != 0)
        {
            candidate[l] = text[l];
            ++l;
            }*/
        uchar *candidate = tc->getSuffix(sapos, patlen+klimit);
        unsigned l = strlen((char*)candidate);
        // If l < patlen, pad candidate with 0's
        while (l < patlen) // FIXME We assume buffer size >= patlen+1
            candidate[l++] = 0;
        candidate[l] = 0;

        if (debug)
            std::cerr << "candidate: " << candidate << std::endl << "candidat2: " << tc->getSuffix(sapos, patlen+klimit) << std::endl << "pattern: " << pat << std::endl;

        std::pair<unsigned, unsigned> dist = med.prefixDist(candidate, l);
        if (dist.first > klimit) {
            delete [] candidate;
            return;
        }

        newMatch(sapos, (char const *)candidate, dist.second);
        delete [] candidate;
    }


    inline void newMatch(ulong sapos, const char *text, unsigned l) 
    {
        this->sum ++;
        TextCollection::position_result pos;
        tc->getPosition(pos, sapos);
        outputw.reportIndels(*p, pos, tc->getName(pos.first), std::string(text, l), 1); // FIXME
    }


    void calculateOptimalPartition(void) 
    {
   
        unsigned i,j;
        unsigned int **Cost;
        Cost = new unsigned int*[patlen];
        for (i=0;i<patlen; i++)
            Cost[i] = new unsigned int[klimit+1];
        // init max cost   
        for (i=0; i<patlen; i++)
            for (j=0; j<klimit+1; j++)
                Cost[i][j] = textlen*(klimit+1);
        unsigned int **Count;
        Count = new unsigned int*[patlen];
        for (i=0;i<patlen; i++)
            Count[i] = new unsigned int[patlen];   
        // init min count   
        for (i=0; i<patlen; i++)
            for (j=0; j<patlen; j++)
                Count[i][j] = 0;

        for (i=patlen; i>0;) {
            --i;
            j=i+1;
            while (j > 0 && pushChar(pat[j-1])) { 
                --j;  	   
                Count[j][i] = smax.top()-smin.top()+1;
            }
            // Reset
            match.clear();
            while (!smin.empty()) smin.pop();
            while (!smax.empty()) smax.pop();
            smin.push(0);
            smax.push(textlen-1);  
        }
        // DP for computing best splitting
        unsigned k;
        for (i=0;i<patlen;i++)
            Cost[i][0] = Count[0][i];    

        for (j=1; j<klimit+1; j++)
            for (i=1;i<patlen;i++)
                for (k=0; k<i; k++)
                    if (Cost[k][j-1]+(unsigned)(Count[k+1][i]*std::min(1.0,choose(k+1,j)*std::pow(3.0,(int)j)*std::pow(0.25,std::max(0,(int)(k+1-j)))))<Cost[i][j])
                        Cost[i][j] = Cost[k][j-1]+(unsigned)(Count[k+1][i]*std::min(1.0,choose(k+1,j)*std::pow(3.0,(int)j)*std::pow(0.25,std::max(0,(int)(k+1-j)))));

        if (splitPos!=0)
            delete [] splitPos;
        splitPos = new unsigned[klimit+2];   
        splitPos[klimit+1] = patlen;
        // trace back
        i=patlen-1;
        if (debug) std::cout << "Estimated number of candidates: " << Cost[i][klimit] << "\n";
        if (debug) std::cout << "Split positions: ";
        for (j=klimit; j>0; j--)
            for (k=0; k<i; k++)
                if (Cost[k][j-1]+(unsigned)(Count[k+1][i]*std::min(1.0,choose(k+1,j)*std::pow(3.0,(int)j)*std::pow(0.25,std::max(0,(int)(k+1-j)))))==Cost[i][j]) {
                    splitPos[j]=k+1;
                    if (debug) std::cout  << k+1 << ",";
                    i = k;
                    break;
                }
        splitPos[0]=0;      
        if (debug) std::cout << "\n";	 
        for (i=0;i<patlen; i++) {
            delete [] Cost[i];
            delete [] Count[i];
        }
        delete [] Cost;
        delete [] Count;
   
        // uniform partition   
        /*
          psize = patlen / (klimit+1);
          splitPos[0]=0;
          for (int i= 1;i<=klimit; i++)
          splitPos[i]=splitPos[i-1]+psize;
          splitPos[klimit+1]=patlen;
        */
    }

    double choose(int n, int m) 
    {        
        int i;
        double result = 1.0;
        if (m > n/2) m = n-m;
        for (i=0;i<m;i++)
            result *= 1.0*(n-i)/(m-i);
        return result;
    }
    
    ResultSet *results;
    unsigned psize;
    unsigned *splitPos;
    MyersEditDistanceIncremental *ed;

    IndelQuery();
    // No copy constructor or assignment
    IndelQuery(IndelQuery const&);
    IndelQuery& operator = (IndelQuery const&);
};

#endif // _IndelQuery_H_
