/**
 * k-mismatch queries
 * 
 * Aligment using case analysis pruning, see README for references.
 */

#ifndef _MismatchQuery_H_
#define _MismatchQuery_H_

#include "Pattern.h"
#include "InputReader.h"
#include "OutputWriter.h"
#include "Query.h"

#include <string>
#include <cstdlib> // exit()

class MismatchQuery : public Query
{
public:
     MismatchQuery(TextCollection *tc, TextCollection *tcr_, OutputWriter &ow, bool vrb, 
                   bool incr, bool rvrs, unsigned rprt)
         : Query(tc, ow, vrb, incr, rvrs, rprt), tcr(tcr_), reversed(false), psize(0),
        piecek(0), pcount(0)
    { }

protected:
    void firstStep();

private:
    TextCollection *tcr; // Index for reversed text
    bool reversed;       // true if tc has been replaced with tcr and tcr with tc
   
    unsigned psize;
    int *piecek;
    unsigned pcount;

    void distribute(unsigned, int, int);

    inline void newMatch() 
    {
        ulong min = smin.top();
        ulong occs = smax.top() - min + 1;
        if (occs > report - sum)
            occs = report - sum;
        this->sum += occs;

        TextCollection::position_vector posv;
        tc->getPosition(posv, min, min + occs - 1);

        if (debug)
            std::cerr << "occurrence(s) found using " << (reversed ? "reversed" : "forward") << " index:" << std::endl;

        // Reconstruct text from stack
        std::string text;
        text.reserve(patlen+1);

        // If using reversed index, pattern and text are backwards
        if (reversed) 
        {
            p->reverse();
            for (unsigned i = 0; i < match.size(); ++i)
                text += match[i];
            
            for (TextCollection::position_vector::iterator pos = posv.begin(); 
                 pos != posv.end(); ++pos) 
            {
                (*pos).second = tc->getLength((*pos).first) - (*pos).second - patlen;
                outputw.reportMismatches(*p, *pos, tcr->getName((*pos).first), text, 
                                         smax.top() - min + 1);
            }
            p->reverse(); // Restore orientation
        }
        else
        {
            // Forward orientation
            for (unsigned i = match.size(); i > 0;)
                text += match[--i];
            
            for (TextCollection::position_vector::iterator pos = posv.begin(); 
                 pos != posv.end(); ++pos) 
                outputw.reportMismatches(*p, *pos, tc->getName((*pos).first), text, 
                                         smax.top() - min + 1);
        }
    }

    inline void swap()
    {
        // Swap tc and tcr to search forward/reversed index
        reversed = !reversed;
        TextCollection * const tmp = tc;
        tc = tcr;
        tcr = tmp;
    }

    inline bool useForwardIndex()
    {
        // Forward index will match the pattern backwards
        for (unsigned i = 0; i < pcount/2; ++i)
            if (piecek[i] < piecek[pcount-i-1])
                return false;
        return true;            
    }

    template<class array_type>
    inline void revarr(array_type * str, unsigned len) 
    {
        array_type t;
        for (unsigned i = 0; i < len / 2; i++) 
        {
            t = str[i];
            str[i] = str[len - i - 1];
            str[len - i - 1] = t;
        }
    }
    
    void firstPiece();
    void nextPiece(int, int, int);
    void exactMatch(int, int);
};

#endif // _MismatchQuery_H_
